@@ -1,5 +1,5 @@
 package Geo::Address::Formatter;
-$Geo::Address::Formatter::VERSION = '1.9989';
+$Geo::Address::Formatter::VERSION = '1.9990';
 # ABSTRACT: take structured address data and format it according to the various global/country rules
 
 use strict;
@@ -164,9 +164,139 @@ sub _read_configuration {
 
     #say Dumper $self->{abbreviations};
     #say Dumper $self->{country2lang};
+
+    $self->_precompile_rules();
+    $self->_build_code_lookups();
+    $self->_precompile_abbreviations();
+
     return 1;
 }
 
+sub _precompile_rules {
+    my $self = shift;
+
+    foreach my $cc (keys %{$self->{templates}}) {
+        my $tpl = $self->{templates}{$cc};
+        next unless is_hashref($tpl);
+
+        # Pre-compile replace rules
+        if (defined($tpl->{replace})) {
+            my @compiled;
+            foreach my $ra_fromto (@{$tpl->{replace}}) {
+                my $pattern = $ra_fromto->[0];
+                my $replacement = $ra_fromto->[1];
+                my ($comp_name, $exact_match, $re);
+
+                # detect component-specific rules like "state=Some Value"
+                if ($pattern =~ m/^(\w+)=(.+)$/) {
+                    $comp_name = $1;
+                    my $rest = $2;
+                    # try to compile as regex; if it fails, it's an exact match
+                    eval { $re = qr/$rest/i };
+                    if ($@) {
+                        warn "invalid replacement regex '$rest' for $cc, skipping";
+                        next;
+                    }
+                    $exact_match = $rest;
+                } else {
+                    eval { $re = qr/$pattern/i };
+                    if ($@) {
+                        warn "invalid replacement regex '$pattern' for $cc, skipping";
+                        next;
+                    }
+                }
+                push @compiled, {
+                    component   => $comp_name,
+                    exact_match => $exact_match,
+                    re          => $re,
+                    replacement => $replacement,
+                };
+            }
+            $tpl->{_compiled_replace} = \@compiled;
+        }
+
+        # Pre-compile postformat_replace rules
+        if (defined($tpl->{postformat_replace})) {
+            my @compiled;
+            foreach my $ra_fromto (@{$tpl->{postformat_replace}}) {
+                my $pattern = $ra_fromto->[0];
+                my $replacement = $ra_fromto->[1];
+                my $re;
+                eval { $re = qr/$pattern/ };
+                if ($@) {
+                    warn "invalid postformat regex '$pattern' for $cc, skipping";
+                    next;
+                }
+                push @compiled, {
+                    re          => $re,
+                    replacement => $replacement,
+                };
+            }
+            $tpl->{_compiled_postformat} = \@compiled;
+        }
+    }
+    return;
+}
+
+sub _build_code_lookups {
+    my $self = shift;
+
+    foreach my $type (qw(state_codes county_codes)) {
+        my $reverse_key = $type . '_reverse';
+        my $name_key    = $type . '_name';
+        $self->{$reverse_key} = {};
+        $self->{$name_key}    = {};
+
+        my $data = $self->{$type} // next;
+        foreach my $cc (keys %$data) {
+            my $mapping = $data->{$cc};
+            my $rev  = {};
+            my $name = {};
+
+            foreach my $code (keys %$mapping) {
+                my $val = $mapping->{$code};
+                if (is_hashref($val)) {
+                    # hash with default, alt, alt_XX keys
+                    my $default_name = $val->{default};
+                    $name->{$code} = $default_name if defined $default_name;
+                    foreach my $v (values %$val) {
+                        $rev->{uc($v)} = $code;
+                    }
+                } else {
+                    # simple string value
+                    $name->{$code} = $val;
+                    $rev->{uc($val)} = $code;
+                }
+                # code-to-code identity lookup
+                $rev->{$code} = $code;
+            }
+            $self->{$reverse_key}{$cc} = $rev;
+            $self->{$name_key}{$cc}    = $name;
+        }
+    }
+    return;
+}
+
+sub _precompile_abbreviations {
+    my $self = shift;
+    my $abbr = $self->{abbreviations} // return;
+
+    foreach my $lang (keys %$abbr) {
+        foreach my $comp_name (keys %{$abbr->{$lang}}) {
+            my $rh_pairs = $abbr->{$lang}{$comp_name};
+            my @compiled;
+            foreach my $long (keys %$rh_pairs) {
+                push @compiled, {
+                    re    => qr/(^|\s)\Q$long\E\b/,
+                    short => $rh_pairs->{$long},
+                };
+            }
+            $self->{compiled_abbreviations}{$lang}{$comp_name} = \@compiled;
+        }
+    }
+    return;
+}
+
 
 sub final_components {
     my $self = shift;
@@ -324,8 +454,8 @@ sub format_address {
     }
 
     # apply replacements if we have any
-    if (defined($rh_config->{replace})){
-        $self->_apply_replacements($rh_components, $rh_config->{replace});
+    if (defined($rh_config->{_compiled_replace})){
+        $self->_apply_replacements($rh_components, $rh_config->{_compiled_replace});
         if ($debug){
             say STDERR "after applying_replacements applied";
             say STDERR Dumper $rh_components;
@@ -377,8 +507,8 @@ sub format_address {
     $template_text = $self->_replace_template_lambdas($template_text);
 
     # 10. compiled the template
-    my $compiled_template =
-        $THC->compile($template_text, {'numeric_string_as_string' => 1});
+    my $compiled_template = $self->{compiled_template_cache}{$template_text}
+        //= $THC->compile($template_text, {'numeric_string_as_string' => 1});
 
     if ($debug){
         say STDERR "before _render_template";
@@ -395,7 +525,7 @@ sub format_address {
     }
 
     # 12. postformatting
-    $text = $self->_postformat($text, $rh_config->{postformat_replace});
+    $text = $self->_postformat($text, $rh_config->{_compiled_postformat});
 
     # 13. clean again
     $text = $self->_clean($text);
@@ -433,29 +563,25 @@ sub _postformat {
     $text = join(', ', @after_pieces);
 
     # do any country specific rules
-    foreach my $ra_fromto (@$raa_rules) {
-        try {
-            my $regexp = qr/$ra_fromto->[0]/;
-            my $replacement = $ra_fromto->[1];
+    foreach my $rule (@$raa_rules) {
+        my $regexp = $rule->{re};
+        my $replacement = $rule->{replacement};
 
-            # ultra hack to do substitution
-            # limited to $1 and $2, should really be a while loop
-            # doing every substitution
+        # ultra hack to do substitution
+        # limited to $1 and $2, should really be a while loop
+        # doing every substitution
 
-            if ($replacement =~ m/\$\d/) {
-                if ($text =~ m/$regexp/) {
-                    my $tmp1 = $1;
-                    my $tmp2 = $2;
-                    my $tmp3 = $3;
-                    $replacement =~ s/\$1/$tmp1/;
-                    $replacement =~ s/\$2/$tmp2/;
-                    $replacement =~ s/\$3/$tmp3/;
-                }
+        if ($replacement =~ m/\$\d/) {
+            if ($text =~ m/$regexp/) {
+                my $tmp1 = $1;
+                my $tmp2 = $2;
+                my $tmp3 = $3;
+                $replacement =~ s/\$1/$tmp1/;
+                $replacement =~ s/\$2/$tmp2/;
+                $replacement =~ s/\$3/$tmp3/;
             }
-            $text =~ s/$regexp/$replacement/;
-        } catch {
-            warn "invalid replacement: " . join(', ', @$ra_fromto);
-        };
+        }
+        $text =~ s/$regexp/$replacement/;
     }
     return $text;
 }
@@ -633,65 +759,37 @@ sub _add_code {
     # ensure country_code is uppercase as we use it as conf key
     my $cc = uc($rh_components->{country_code});
 
-    if (my $mapping = $self->{$code . 's'}{$cc}) {
+    my $reverse_key = $code . 's_reverse';
+    my $name_key    = $code . 's_name';
 
+    if (my $rev = $self->{$reverse_key}{$cc}) {
         my $name    = $rh_components->{$keyname};
         my $uc_name = uc($name);
 
-        LOCCODE: foreach my $abbrv (keys %$mapping) {
-
-            my @confnames; # can have multiple names for the place
-                           # for example in different languages
-
-            if (is_hashref($mapping->{$abbrv})) {
-                push(@confnames, values %{$mapping->{$abbrv}});
-            } else {
-                push(@confnames, $mapping->{$abbrv});
-            }
-
-            # FIXME: should only uc the names once when reading from conf
-            foreach my $confname (@confnames) {
-
-                if ($uc_name eq uc($confname)) {
-                    $rh_components->{$code} = $abbrv;
-                    last LOCCODE;
-                }
-                # perhaps instead of passing in a name, we passed in a code
-                # example: state => 'NC'
-                # we want to turn that into
-                #     state => 'North Carolina'
-                #     state_code => 'NC'
-                #
-                if ($uc_name eq $abbrv) {
-                    $rh_components->{$keyname} = $confname;
-                    $rh_components->{$code}    = $abbrv;
-                    last LOCCODE;
-                }
+        if (my $found_code = $rev->{$uc_name}) {
+            # matched a name -> code
+            $rh_components->{$code} = $found_code;
+            # if input was a code (e.g. state => 'NC'), set keyname to full name
+            if ($uc_name eq $found_code) {
+                my $full_name = $self->{$name_key}{$cc}{$found_code};
+                $rh_components->{$keyname} = $full_name if defined $full_name;
             }
         }
-        # didn't find a valid code or name
 
-        # try again for odd variants like "United States Virgin Islands"
-        if ($cc eq 'US') {
-            if ($keyname eq 'state') {
-                if (!defined($rh_components->{state_code})) {
-                    if ($rh_components->{state} =~ m/^united states/i) {
-                        my $state = $rh_components->{state};
-                        $state =~ s/^United States/US/i;
-                        foreach my $k (keys %$mapping) {
-                            if (uc($state) eq uc($k)) {
-                                $rh_components->{state_code} = $mapping->{$k};
-                                last;
-                            }
-                        }
-                    }
-                    if ($rh_components->{state} =~ m/^washington,? d\.?c\.?/i) {
-                        $rh_components->{state_code} = 'DC';
-                        $rh_components->{state}      = 'District of Columbia';
-                        $rh_components->{city}       = 'Washington';
-                    }
+        # US-specific edge cases
+        if ($cc eq 'US' && $keyname eq 'state' && !defined($rh_components->{state_code})) {
+            if ($rh_components->{state} =~ m/^united states/i) {
+                my $state = $rh_components->{state};
+                $state =~ s/^United States/US/i;
+                if (my $fc = $rev->{uc($state)}) {
+                    $rh_components->{state_code} = $fc;
                 }
             }
+            if ($rh_components->{state} =~ m/^washington,? d\.?c\.?/i) {
+                $rh_components->{state_code} = 'DC';
+                $rh_components->{state}      = 'District of Columbia';
+                $rh_components->{city}       = 'Washington';
+            }
         }
     }
     return $rh_components->{$code};
@@ -700,38 +798,59 @@ sub _add_code {
 sub _apply_replacements {
     my $self          = shift;
     my $rh_components = shift;
-    my $raa_rules     = shift // return; # bail out if no rules
+    my $ra_rules      = shift // return; # bail out if no rules
 
     if ($debug){
         say STDERR "in _apply_replacements";
-        say STDERR Dumper $raa_rules;
+        say STDERR Dumper $ra_rules;
+    }
+
+    # support legacy raw array-of-arrays format (used by tests)
+    if (ref($ra_rules) eq 'ARRAY' && @$ra_rules && ref($ra_rules->[0]) eq 'ARRAY') {
+        foreach my $component (keys %$rh_components) {
+            next if ($component eq 'country_code');
+            next if ($component eq 'house_number');
+            foreach my $ra_fromto (@$ra_rules) {
+                my $regexp;
+                if ($ra_fromto->[0] =~ m/^$component=/) {
+                    my $from = $ra_fromto->[0];
+                    $from =~ s/^$component=//;
+                    if ($rh_components->{$component} eq $from) {
+                        $rh_components->{$component} = $ra_fromto->[1];
+                    } else {
+                        $regexp = $from;
+                    }
+                } else {
+                    $regexp = $ra_fromto->[0];
+                }
+                if (defined($regexp)) {
+                    try {
+                        my $re = qr/$regexp/i;
+                        $rh_components->{$component} =~ s/$re/$ra_fromto->[1]/;
+                    } catch {
+                        warn "invalid replacement: " . join(', ', @$ra_fromto);
+                    };
+                }
+            }
+        }
+        return $rh_components;
     }
 
     foreach my $component (keys %$rh_components) {
         # some components dont need replacements
         next if ($component eq 'country_code');
         next if ($component eq 'house_number');
-        foreach my $ra_fromto (@$raa_rules) {
-            my $regexp;
-            # do key specific replacement
-            if ($ra_fromto->[0] =~ m/^$component=/){
-                my $from = $ra_fromto->[0];
-                $from =~ s/^$component=//;
-                if ($rh_components->{$component} eq $from){
-                    $rh_components->{$component} = $ra_fromto->[1];
+        foreach my $rule (@$ra_rules) {
+            if (defined($rule->{component})) {
+                # component-specific rule: only applies to matching component
+                next if ($rule->{component} ne $component);
+                if ($rh_components->{$component} eq $rule->{exact_match}) {
+                    $rh_components->{$component} = $rule->{replacement};
                 } else {
-                    $regexp = $from;
+                    $rh_components->{$component} =~ s/$rule->{re}/$rule->{replacement}/;
                 }
             } else {
-                $regexp = $ra_fromto->[0];
-            }
-            if (defined($regexp)){
-                try {
-                    my $re = qr/$regexp/i;
-                    $rh_components->{$component} =~ s/$re/$ra_fromto->[1]/;
-                } catch {
-                    warn "invalid replacement: " . join(', ', @$ra_fromto);
-                };
+                $rh_components->{$component} =~ s/$rule->{re}/$rule->{replacement}/;
             }
         }
     }
@@ -763,16 +882,14 @@ sub _abbreviate {
         my @langs = split(/,/, $self->{country2lang}{$cc});
 
         foreach my $lang (@langs) {
-            # do we have abbrv for this lang?
-            if (defined($self->{abbreviations}->{$lang})) {
-                # we have abbreviations
-                my $rh_abbr = $self->{abbreviations}->{$lang};
+            # do we have compiled abbrv for this lang?
+            if (defined($self->{compiled_abbreviations}{$lang})) {
+                my $rh_compiled = $self->{compiled_abbreviations}{$lang};
 
-                foreach my $comp_name (keys %$rh_abbr) {
+                foreach my $comp_name (keys %$rh_compiled) {
                     next if (!defined($rh_comp->{$comp_name}));
-                    foreach my $long (keys %{$rh_abbr->{$comp_name}}) {
-                        my $short = $rh_abbr->{$comp_name}->{$long};
-                        $rh_comp->{$comp_name} =~ s/(^|\s)$long\b/$1$short/;
+                    foreach my $rule (@{$rh_compiled->{$comp_name}}) {
+                        $rh_comp->{$comp_name} =~ s/$rule->{re}/$1$rule->{short}/;
                     }
                 }
             } else {
@@ -848,7 +965,7 @@ sub _render_template {
     my $components = shift;
 
     # Mustache calls it context
-    my $context = clone($components);
+    my $context = $components;
     say STDERR 'context: ' . Dumper $context if ($debug);
     my $output = $thtemplate->render($context);
 
@@ -962,6 +1079,7 @@ sub _find_unknown_components {
     return \@a_unknown;
 }
 
+
 1;
 
 __END__
@@ -976,7 +1094,7 @@ Geo::Address::Formatter - take structured address data and format it according t
 
 =head1 VERSION
 
-version 1.9989
+version 1.9990
 
 =head1 SYNOPSIS
 
@@ -1002,40 +1120,64 @@ Returns one instance. The I<conf_path> is required.
 
 Optional parameters are:
 
-I<debug>: prints tons of debugging info for use in development.
+=over
 
-I<no_warnings>: turns off a few warnings if configuration is not optimal.
+=item * I<debug>
 
-I<only_address>: formatted will only contain known components (will not include POI names). Note, can be overridden with optional param to format_address method.
+Prints tons of debugging info for use in development.
+
+=item * I<no_warnings>
+
+Turns off a few warnings if configuration is not optimal.
+
+=item * I<only_address>
+
+Formatted output will only contain known address components (will not
+include POI names). Can be overridden per-call via the I<only_address>
+option to L</format_address>.
+
+=back
 
 =head2 final_components
 
   my $rh_components = $GAF->final_components();
 
-returns a reference to a hash of the final components that are set at the
-completion of B<format_address>. Warns if called before they have been set 
-(unless I<no_warnings> was set at object creation).
+Returns a reference to a hash of the final components that were set during
+the most recent call to B<format_address>.  Returns C<undef> if called
+before B<format_address> has been called.  Warns in that case unless
+I<no_warnings> was set at object creation.
 
 =head2 format_address
 
   my $text = $GAF->format_address(\%components, \%options );
 
 Given structured address components (hashref) and options (hashref) returns a
-formatted address (string).
+formatted address as a multiline string with a trailing newline.
 
 Possible options are:
 
-    'abbreviate', if supplied common abbreviations are applied
-    to the resulting output.
+=over
 
-    'address_template', a mustache format template to be used instead of the template
-    defined in the configuration
+=item * I<abbreviate>
 
-    'country', which should be an uppercase ISO 3166-1:alpha-2 code
-    e.g. 'GB' for Great Britain, 'DE' for Germany, etc.
-    If omitted we try to find the country in the address components.
+If true, common abbreviations are applied to the resulting output.
 
-    'only_address', same as only_address global option but set at formatting level
+=item * I<address_template>
+
+A Mustache-format template to be used instead of the template defined in
+the configuration.
+
+=item * I<country>
+
+An uppercase ISO 3166-1 alpha-2 code, e.g. C<'GB'> for Great Britain,
+C<'DE'> for Germany, etc.  If omitted the country is determined from
+the address components.
+
+=item * I<only_address>
+
+Same as the I<only_address> constructor option but applied per-call.
+
+=back
 
 =head1 DESCRIPTION
 
@@ -1068,13 +1210,20 @@ which includes test cases.
 
 Together we can address the world!
 
+=head1 SEE ALSO
+
+L<https://github.com/OpenCageData/address-formatting> - the address
+formatting template configuration used by this module.
+
+L<https://opencagedata.com> - OpenCage geocoder.
+
 =head1 AUTHOR
 
 Ed Freyfogle
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2025 by Opencage GmbH.
+This software is copyright (c) 2026 by Opencage GmbH.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
